#=======================================================
# Created by Caitlin C. Bannan
# Graduate Student
# Mobley Lab
# University of California, Irvine
# July 2015
#=======================================================

import numpy as np      # Don't you always need it?
import commands as c    # work in command line 
import pickle
import pymbar           # used for statistical inefficiency
import os               # used to check input file exists
import sys              # used to exit if errors are made
from optparse import OptionParser
import matplotlib.pyplot as plt
from pylab import *

#=======================================================
# Methods for convergence data
#=======================================================

def getOutput(inputFN, output, prop, start_time = 0.0, end_time = -1.0):
    """
    This method gets an output xvg file using GROMACS g_energy. 

    Input:
        inputFN - input file name, string ending in .edr
        output - output file naem, string ending in .xvg
        prop - property you want reported in the xvg file, string
        start_time - optional, number of picoseconds to skip at beginning of data, float (no data removed if = 0)
        end_time - optional, number of picoseconds to skip at end of data, float (no data removed if = -1)

    output:
        Makes a GROMACS .xvg file with the data for the property (prop) provided
    """
    #error if the input file isn't in the current path
    if os.path.isfile(inputFN) == False:
        fileError = Exception("input file (%s) not found" % inputFN)
        raise fileError

    #use g_energy to get .xvg file
    string = "echo %s | g_energy -f %s -o %s" % (prop, inputFN, output)
    
    if start_time != 0:
        string = string +' -b '+str(start_time)
    if end_time != -1:
        string = string+' -e '+str(end_time)
    c.getoutput(string)
    print "Gromacs used to get .xvg file for %s, can be found at %s" %(prop, output)

def makeArray(output):
    """
    Input:
        output - output file name, string ending in .xvg

    Output:
        returns an array with all data from the .xvg file 
    """
    #extract data from  .xvg file
    allData = np.genfromtxt(output, skip_header=22)
    print "Data collected from .xvg file can be found in your dictionary at 'data' as an array with (time, property)"
    return allData 

def getConvergenceData(data, percent):
    """
    Makes convergence data
    Input:
        data - same format as created from the xvg files with rows of data where the columns are x = time and y = property
        percent - how frequently to "cut" data to make a point on the convergence chart, so 1% would make 100 points for example. 

    Output:
        list of data for 
        x = time at each average taken
        y = average from begining to that point
        err = err in average value and is the standard deviation divided by the squareroot of the effective number of samples. 
    """
    x = []
    y = []
    err = []
    xs = data[:,0]
    percent = percent / 100.
    cut = int(percent * len(data))
    for i in range(1, len(data)):
        if i % cut == 0 or i == (len(data)-1):
            x.append(xs[i])
            ys = data[0:i,1]
            y.append(np.average(ys))
            g = pymbar.timeseries.statisticalInefficiency(ys)
            N = i+1
            effSample = np.sqrt(N/g)
            err.append(np.std(ys)/effSample)
    print "Convergence data can be found in your dictionary at 'conv_data'"        
    return np.array(x), np.array(y), np.array(err)

def makePlots(x, y, err, prop, units, fig, plot_name, prop_expected = None):
    """
    Makes a plot given convergence data and titles. 
    Input:
        x = 'time' data, array or list
        y = average property data, array or list
        err = error in average property data, array or list
        prop = property being plotted, string
        units = units for the property being plotted, string
        fig = a pyplot or other figure that an axis can be added to. 
        plot_name = name for file created, type is default png, but will be taken from extension, string
        prop_expected = where you expect the property to converge, default = None will use the last average, otherwise float

    Output:
        returns figure that was created
        Creates a file with name plot_name
    """
    ax=fig.add_subplot(111)
    ax.set_xlabel("time (ps)")
    ax.set_ylabel("%s (%s)" %(prop, units))
    ax.set_title("%s Convergence Plot" %prop)    
    ax.set_xlim(0,x[-1]+1)
    ax.set_ylim(min(y)-5, max(y)+5)
    ax.errorbar(x, y, yerr=err, fmt='bo', label = 'average %s'%prop)

    if prop_expected is None:
        yExp = y[-1]
        ax.plot([-1, (max(x)+10)],[yExp,yExp], 'k-', label='average %.1f'%yExp)
    else:
        yExp = prop_expected
        yAve = y[-1]
        ax.plot([-1, (max(x)+10)],[yExp,yExp], 'k-', label='expected %.1f (average %.1f'%(yExp, yAve))
    ax.legend(bbox_to_anchor=(0.5, -0.2), loc = "upper center", borderaxespad = 0., ncol = 2, fancybox = True)

    fig.savefig(plot_name, dpi = 600, bbox_inches = 'tight', pad_inches = 0.5, orientation = 'landscape')
    print "Convergence plot created, it can be found at %s and in your dictionary at 'fig'" % plot_name
    return fig

# Use the other methods together to make convergence data and plot
def Converge(prop, makeOutput = True, inputFile = 'input.edr', output = 'output.xvg', plot_name = 'ConvergePlot.pdf', start_time = 0.0, end_time = -1., percent = 2., prop_expected = None, saveData = False, dictName = "converge_data.p"):
    """
    INPUT: 
        prop = STRING for the property you would like to monitor, the options are:
            Volume, Pressure, Potential, pV, Kinetic-En, Density, Total-Energy, Enthalpy, Temperature 
        makeOutput = (True) BOOLEAN: False you must provide an output that already exists and has the data for the property provided
        inputFile = ('input.edr') STRING path to GROMACS energy file (.edr), no need to provide if makeOutput is False 
        output = ('output.xvg') STRING path for the GROMACS output file (.xvg), if makeOutput is True then this is where the output will be stored, if it's False than this should be where your data file is already stored.  
        plot_name = ('ConvergePlot.pdf') STRING:name for the figure file, type detected by extension 
        start_time = (0.0) FLOAT: time to start collecting data (ps) 
        end_time = (-1.) FLOAT: time to stop collecting data (ps) (if -1 then all data is used) 
        percent = 2.,
        expected = (None) FLOAT, expected value for property, this will be a horizontal line on the convergence plot, the average of all the data is used if this isn't provided.
        saveData = (False) BOOLEAN, if True a pickle file will be created with a dictionary of all of the input information and calculated data
        dictName = ('converge_data.p') STRING name for pickle file of data, it is only created if saveData = True

    OUTPUT:
        Creates an image file of the convergence data for the provided property (prop) and returns a matplotlib figure and arrays with time, average property, and uncertainty in average property used to make this plot. 

    There are also two optional output files:
        1. It can create the GROMACS .xvg file for the property if you don't already have itwith the provided name (output) or the default output.xvg. 
        2. A dictionary file that contains information about the data stored as a pickle file. The dictionary has the following keys, corresponding to provided and calculated information: 'input', 'output', 'property', 'start_time', 'end_time', 'expected', 'units', 'data', 'conv_data': dictionary - of np arrays for the keys: 'x' (or time), 'y' (or average property), 'err' (error in the average property)
    """
    # Used to determine if the property provided can be claculated with Gromacs, then used to assign units.
    propDict = {'Volume': 'nm^3','Pressure': 'bar', 'Potential': 'kJ/mol','pV':'kJ/mol', 'Kinetic-En': 'kJ/mol', 'Density': 'kg/m^3','Total-Energy':'kJ/mol', 'Enthalpy': 'kJ/mol','Temperature':'K'}
    
    # exit if invalid property is provided and you need to make output
    try:
        unit = propDict[prop]
    except:
        if makeOutput:
            propertyError = Exception("the property you provided is not available, please choose from the list\n%s" % str(','.join(propDict.keys())))
            raise propertyError
        else:
            print "Property not found so units not assigned, however data from the .xvg file will still be analyzed."
            unit = ''

    if makeOutput:
        # get output file
        getOutput(inputFile, output, prop, start_time, end_time)

    # get data array and save to dictionary
    data = makeArray(output)

    # get convergence data and save to dictionary inside D
    x, y, err = getConvergenceData(data, percent)

    # Make figure, make plot and save figure to dictionary
    fig = figure(1)
    fig = makePlots(x, y, err, prop, unit, fig, plot_name, prop_expected)

    # If saveData make dictionary and output as pickle file
    if saveData:
        D = {'input': inputFile, 'output': output,'property': prop, 'start_time': start_time, 'end_time': end_time, 'plot_name': plot_name, 'percent': percent, 'prop_expected': expected, 'units': propDict[prop], 'conv_data': {'x':x,'y': y,'err': err}, 'data': np.array(data)}
        pickle.dump(D, open(dictName,'wb'))
        print "All convergence related data is saved in pickled dictionary 'converge_data.p' and your convergence plot was saved to %s" % plot_name 

    #Return figure
    return fig

# =======================================================
# Running from Commandline
# =======================================================

if __name__ == '__main__':
    # List of GROMACS available properties
    propList = ['Volume','Pressure', 'Potential','pV', 'Kinetic-En', 'Density','Total-Energy','Enthalpy','Temperature']
 
    par = OptionParser()
    #=======================================================
    # Use option parser to get information from input
    #=======================================================

    par.add_option('-p',
            '--property', 
            dest = 'prop', 
            help = """string - MANDATORY (no default)
            Property you want a convergence plot for. The options are %s
            """ % str(', '.join(propList)))

    par.add_option('-f',
            '--input', 
            dest = 'inputFN', 
            help = """string - Default = input.edr 
            path to gromacs energy file (.edr)
            """,
            default = 'input.edr')

    par.add_option('-o',
            '--output', 
            dest = 'output', 
            help = """string - Default = output.xvg
            path for the gromacs output file (.xvg)  
            """,
            default = 'output.xvg')

    par.add_option('-N',
            '--plot_name', 
            dest = 'plot_name', 
            help = """string - Default = ConvergePlot.pdf
            name for the figure file, type detected by extension
            """,
            default = 'ConvergePlot.pdf')

    par.add_option('-b',
            '--start_time', 
            dest = 'start_time', 
            help = """float - Default = 0.
            specify time in ps to start collecting data
            """,
            default = '0.',
            type = 'float')

    par.add_option('-e',
            '--end_time', 
            dest = 'end_time',
            help = """float - Default = -1. (till end)
            specify time in ps to stop collecting data
            """,
            default = '-1.',
            type = 'float')

    par.add_option('-%',
            '--percent', 
            dest = 'percent', 
            help = """float - Default = 2%
            how frequently to average data as a percentage of the total
            """,
            default = '2.',
            type = 'float')

    par.add_option('-x',
            '--expected', 
            dest = 'prop_expected', 
            help = """float - Default (None)
            If known, what is the expected value for the property
            """,
            type = 'float')

    par.add_option('-m', '--makeOutput',
            dest = 'makeOutput',
            help = "boolean - Should this script make the output xvg file or does it already exist",
            default = 'True')

    par.add_option('-d','--dictName',
            dest = 'dictName',
            help = "String - Name of the pickle file for the output dictionary, should end in .p or .pickle, if this is not provided then the data will not be saved to a pickle file and only the Convergence figure will be created.")

    P = par.parse_args()[0]

    # exit if property isn't provided
    if P.prop is None:
        par.error("ERROR: You must provide a property, specify with -p. You can choose from the following: \n%s" % str('\n'.join(propDict.keys())))

    try:
        makeOutput = bool(P.makeOutput)
    except:
        par.error('Do you already have an .xvg file with data? -m or --makeOutput needs to be a valid boolean please try again')

    if makeOutput and not P.prop in propList:
        par.error("ERROR: The property (specified with -p) should be one of the following: \n%s" % str('\n'.join(propDict.keys())))


    if P.dictName == None:
        saveData = False
        
    Converge(P.prop, makeOutput, P.inputFN, P.output, P.plot_name, P.start_time, P.end_time, P.percent, P.prop_expected, saveData, P.dictName) 
