"""
Use the openmoltools wrappers of OpenEye and Antechamber to rebuild input files
Outputs two LOCAL directories of files: ./tripos_mol2/ and ./mol2files_gaff/

Adapted by Guilherme Duarte Ramos Matos, May 2015 
Adapted again by Daisy Y. Kyu August 2015
Some comments and method documentation in triple quotes added by Caitlin C. Bannan in April 2016
Note from Caitlin: Guilherme thinks David Mobley wrote the original script, we're not sure when. 

"""
from MDPandRunMethods import *
import shutil
import pickle
import os
import subprocess
import glob
import openmoltools
import sys 
import commands
path = os.getcwd()

# load Dictionary
SOLV_FILENAME = path+'/dictionary_roundTwo.p'
database = pickle.load(open(SOLV_FILENAME))
#splits full pathname and returns path and filename

def make_path(filename):
    """
    Uses os to make the path for the provided filename.
    """
    try:
        path = os.path.split(filename)[0]
        os.makedirs(path)
    except OSError:
        pass

#makes directory and subdirectory, moves .top and .gro files
def file_organizer(directory, subdirectory, filename):
    """
    uses os to move into directory and create subdirectory. Then moves original topology or coordinate file name to subdirectory and renames it. 
    """
    try:
        os.chdir( directory )
        os.mkdir( subdirectory )
        subprocess.call('mv '+filename+'.gro '+subdirectory+'/morph.gro', shell=True)
        subprocess.call('mv '+filename+'.top '+subdirectory+'/morph.top', shell=True)
        subprocess.call('cp ../New_MDP_files/* '+subdirectory+'/', shell=True)
        os.chdir( subdirectory )
        # Removes molecule name from topology file and puts MOL in place
        try:
            top = open('morph.top','r+')
            inter_top = top.readlines()
            top.close()
        except:
            raise NameError('No morph.top file in %s' % subdirectory)

        #substitution 1
        string1, subst1 = ';name            nrexcl', ' MOL     3\n'
        idx = 0
        while string1 not in inter_top[idx]:
            idx += 1
        inter_top[idx+1] = subst1

        #substitution 2
        string2, subst2 = '; Compound        nmols', 'MOL                 1\n'
        idx = 0
        while string2 not in inter_top[idx]:
            idx += 1
        inter_top[idx+1] = subst2

        #rewrites topology file
        try:
            file = open('morph.top','w+')
            file.writelines( inter_top )
            file.close()
        except:
            raise NameError('No morph.top file in %s' % subdirectory)
        os.chdir('../..')
    except OSError:
        pass
    return

# Set numbers that are alcohols and will need GAFF-DC
alcohol=['1307', '1425', '1453', '159', '2185', '2323', '2336', '2348', '2790', '2914', '2917', '2925', '2938', '5785', '835', '950']

#for each molecule in database, combine paths into tripos_mol2, mol2files_gaff, topgro, topgrosolv
for (key, entry) in database.items():
    # Make all the directories and subdirectories. 
    print "Processing molecule %s ..." % (key)
    tripos_filename = os.path.join('tripos_mol2', entry['number'] + '.mol2')
    frcmod_filename = os.path.join("./mol2files_gaff/", "%s.frcmod" % entry['number'])
    gaff_mol2_filename = os.path.join("./mol2files_gaff/", "%s.mol2" % entry['number'])
    prmtop_filename = os.path.join("./mol2files_gaff/", "%s.prmtop" % entry['number'])
    inpcrd_filename = os.path.join("./mol2files_gaff/", "%s.inpcrd" % entry['number'])
    gro_filename = os.path.join("./topgro/", "%s.gro" % entry['number'])
    top_filename = os.path.join("./topgro/", "%s.top" % entry['number'])
    gro_solv_filename = os.path.join("./topgrosolv/", "%s.gro" % entry['number'])
    top_solv_filename = os.path.join("./topgrosolv/", "%s.top" % entry['number'])
    
    #Add output directories if not already present
    make_path( 'tripos_mol2/' )
    make_path( 'mol2files_gaff/')
    make_path( 'topgro/')
    make_path( 'topgrosolv/')
    #makes the molecule out of smile
    molecule = openmoltools.openeye.smiles_to_oemol(entry['smile'])
    # Assign AM1-BCC charges
    charged = openmoltools.openeye.get_charges(molecule)
    openmoltools.openeye.molecule_to_mol2(charged, tripos_filename)
    #Use openmoltools to make monomer topology files
    _, _ = openmoltools.utils.run_antechamber(key, tripos_filename, charge_method=None, gaff_mol2_filename=gaff_mol2_filename, frcmod_filename=frcmod_filename)
    openmoltools.utils.run_tleap(key, gaff_mol2_filename, frcmod_filename, prmtop_filename=prmtop_filename, inpcrd_filename=inpcrd_filename)
    openmoltools.utils.convert_via_acpype( key, prmtop_filename, inpcrd_filename, out_top = top_filename, out_gro = gro_filename)
    # Use openmoltools gromacs module to solvate with Tip3p water
    openmoltools.gromacs.do_solvate(top_filename, gro_filename, top_solv_filename, gro_solv_filename, water_top='tip3p.itp', box_dim=1.0, box_type='cubic', water_model='spc216.gro')

    # use numbers instead of molecules for file names
    num = entry['number']

    # Use file_organizer to extract 
    file_organizer('topgrosolv', num, num)
    
    # If num is in the alcohol list use Hydroxynator 
    # (Note from Caitlin: This was the old version of hydroxynator since the topology files only had alcohol solutes and no 'oh' atoms in the solvent we never reran the water set up with the new Hydroxynator, in practice you would get the same topology file either way). 
    if num in alcohol:
        os.mkdir('./topgrosolv/%s_OH' % num)
        # Copy the GAFF topology and coordinate files
        commands.getoutput('cp ./topgrosolv/%s/* /work/cluster/kyud/waterSolv/topgrosolv/%s_OH/' % (num, num))
        os.chdir('./topgrosolv/%s_OH' % num)
        # Use Hydroxynator
        commands.getoutput('python /work/cluster/kyud/waterSolv/hydroxynator.py morph.top')
        # Use methods from MDPandRunMethods to make 20 copies of the mdp and run scripts
        make_runFiles(20, 'run.X.sh', num+'_OH')
        make_MDPFiles(20, "XXX")
        os.chdir('../../')     
        
    # For all compounds use methods from MDPandRunMethods to make 20 copies of the mdp and run scripts
    os.chdir('./topgrosolv/%s' % num)
    make_runFiles(20, 'run.X.sh', num)
    make_MDPFiles(20, "XXX")
    os.chdir('../../')
